from .store_file import store_file
from .load_file import load_file, load_file_info
from .store_file_local import store_file_local
from .link_file import link_file
from .cat_file import cat_file
from .core import store_text, store_json, store_npy, store_pkl
from .core import load_text, load_json, load_npy, load_pkl
from .core import store_text_local, store_json_local, store_npy_local, store_pkl_local
from .load_bytes import load_bytes
from .get_kachery_cloud_dir import get_kachery_cloud_dir, use_sandbox
from .init import init
from .get_client_id import get_client_id
from .mutable_local import delete_mutable_folder_local, delete_mutable_local, get_mutable_local, set_mutable_local
from .TemporaryDirectory import TemporaryDirectory
from ._sha1_of_dict import sha1_of_dict
from .zenodo_upload.zenodo_upload import initiate_zenodo_upload
from .request_file import request_file


from typing import Any
import pickle

def _safe_pickle(fname: str, x: Any):
    _check_safe_for_pickling(x)
    with open(fname, 'wb') as f:
        pickle.dump(x, f)

safe_builtins = {
    'range',
    'complex',
    'set',
    'frozenset',
    'slice',
}

class RestrictedUnpickler(pickle.Unpickler):
    def find_class(self, module, name):
        # carefully limit which classes may be loaded
        okay = False
        if module == "builtins" and name in safe_builtins:
            okay = True
        elif module == 'numpy':
            if name in ['ndarray', 'dtype']:
                okay = True
        elif module == 'numpy.core.multiarray':
            if name in ['_reconstruct', 'scalar']:
                okay = True
        if okay:
            return pickle.Unpickler.find_class(self, module, name)
        else:
            raise Exception(f'Not able to safe-unpickle {module}/{name}. If this is safe, consider whitelisting this module/name.')

def _safe_unpickle(fname: str):
    with open(fname, 'rb') as f:
        return RestrictedUnpickler(f).load()

def _check_safe_for_pickling(x: Any):
    if isinstance(x, int) or isinstance(x, float) or isinstance(x, str) or isinstance(x, bool) or (x is None):
        pass
    elif isinstance(x, range) or isinstance(x, complex) or isinstance(x, slice):
        # do not include "set" for now
        pass
    elif isinstance(x, dict):
        y = {}
        for k, v in x.items():
            _check_safe_for_pickling(v)
    elif isinstance(x, list):
        for a in x:
            _check_safe_for_pickling(a)
    elif isinstance(x, tuple):
        for a in x:
            _check_safe_for_pickling(a)
    elif _is_numpy_array(x):
        pass
    elif _is_numpy_number(x):
        pass
    elif _is_numpy_bool(x):
        pass
    else:
        raise Exception(f'Not safe for pickling: ({type(x)}). Perhaps this type should be whitelisted.')

def _is_numpy_array(x):
    try:
        import numpy as np
    except:
        return False
    return isinstance(x, np.ndarray)

def _is_numpy_number(x):
    try:
        import numpy as np
    except:
        return False
    return isinstance(x, np.integer) or isinstance(x, np.floating) or isinstance(x, np.complexfloating)

def _is_numpy_bool(x):
    try:
        import numpy as np
    except:
        return False
    return isinstance(x, np.bool) or isinstance(x, np.bool_)


import os

_kachery_zone_info_by_zone = {}

_default_kachery_gateway_url = 'https://kachery-gateway.figurl.org'

def _get_kachery_zone_info(kachery_zone: str):
    if kachery_zone in _kachery_zone_info_by_zone:
        return _kachery_zone_info_by_zone[kachery_zone]
    from ._kachery_gateway_request import _kachery_gateway_request
    resp = _kachery_gateway_request({
        'type': 'getZoneInfo',
        'zoneName': kachery_zone
    }, kachery_gateway_url=_default_kachery_gateway_url) # important to use default here so we don't get an infinite recursion
    _kachery_zone_info_by_zone[kachery_zone] = resp
    return resp

def _get_kachery_gateway_url():
    url = os.environ.get('KACHERY_GATEWAY_URL', None)
    if url is not None:
        return url
    kachery_zone = os.environ.get('KACHERY_ZONE', None)
    if kachery_zone is not None:
        resp = _get_kachery_zone_info(kachery_zone)
        if resp['found']:
            return resp['kacheryGatewayUrl']
        else:
            raise Exception(f'Unrecognized kachery zone: {kachery_zone}')
    return _default_kachery_gateway_url